@@ -19,6 +19,11 @@ class SyncRequest(JsonBaseRequest):
     contents: str
 
 
+@dataclass
+class GetStatusRequest(JsonBaseRequest):
+    pass
+
+
 @dataclass
 class FocusCellRequest(JsonBaseRequest):
     cell_index: int
